#!/usr/bin/env/python2

################################################################################

# Functions in this module provide the mouse tibiae data.

################################################################################

import sys
import os
from random import choice
import stat

import pydicom
import numpy as np
import keras.utils as np_utils
from skimage.feature import register_translation

################################################################################

mouse_types             = ["Anabolic", "WildType"]
mouse_ids               = [1, 2, 3, 4, 5]
test_mouse_id           = 1
weeks                   = [0, 2, 3, 4, 5, 6, 7, 8]
height, width, channels = 501, 763, 1
batch_size              = 32
train_global            = None
val_global              = None
test_global             = None
num_train_global        = 0
num_val_global          = 0
num_test_global         = 0

################################################################################

def mouse_type_to_label(t):
    """Convert from human-readable treatment class to numeric.

    Args:
        t (str): Treatment class.
            Valid values are: 'Anabolic', 'Ovarectomy', 'WildType'.

    Returns:
        int: Numeric identifier for the treatment class.

    Raises:
        KeyError: If the requested treatment class `t` is not recognised.

    """
    if t not in mouse_types:
        print("\nMouse type not found\n")
        raise KeyError
    return mouse_types.index(t)


def get_dir_name(mouse_type, mouse, week):
    """Find directory containing microCT image data.

    This is dependent on the structure in the filesystem,
    and assumes the script shares a common parent with a directory
    called 'Images', containing the mouse image data in subdirectories.

    Args:
        mouse_type (str): Treatment class.
        mouse (int): Identifier for a specific mouse.
        week (int): Week in which microCT images were collected.

    Returns:
        str: Directory name containing image data, should it exist.

    """
    basedir = os.getcwd()
    while os.path.split(basedir)[1] != "Images":
        if "Images" in os.listdir(basedir):
            basedir = os.path.join(basedir, "Images")
            break
        basedir = os.path.split(basedir)[0]
    dirname = ''.join(['M', str(mouse), 'W', str(week)])
    dirname = os.path.join(basedir, mouse_type, dirname)
    return dirname


def get_max_dim():
    """Calculate maximum height, width, and depth of image dataset.

    The calculated height, width, and depth values are both returned and
    stored as global variables in the module.

    This method should be called only once - thereafter, `get_hwd()`
    should be preferred as it does not perform re-calculation.

    Args:
        None

    Returns:
        (int, int, int): Height, width, depth

    """
    max_rows = 0
    max_cols = 0
    max_channels = 1

    for t in mouse_types:
        for m in mouse_ids:
            for w in weeks:
                dirname = get_dir_name(t, m, w)
                if not os.path.exists(dirname):
                    continue
                img_names = os.listdir(dirname)
                os.chdir(dirname)
                for i in img_names:
                    header = pydicom.read_file(i)
                    rows = header.Rows
                    cols = header.Columns
                    chans = header.SamplesPerPixel
                    if rows > max_rows:
                        max_rows = rows
                    if cols > max_cols:
                        max_cols = cols
                    if chans > max_channels:
                        max_channels = chans

    os.chdir(os.path.join(os.pardir, os.pardir))
    global height
    global width
    global channels
    height, width, channels = max_rows, max_cols, max_channels
    return (max_rows, max_cols, max_channels)


def get_hwd():
    """Return maximum height, width, and depth of image dataset.

    This should be used in preference to `get_max_dim()` after that has
    been called once.

    Args:
        None

    Returns:
        (int, int, int): Height, width, depth.

    """
    global height
    global width
    global channels
    return (height, width, channels)


def resize_helper(old_pixels, min_pixel, dim, max_dim=None, dtype="float32"):
    """Return array of pixels containing the input array centred.

    The output 2D array has dimensions determined by either `max_dim` or
    the maximum dimensions calulated from the image dataset.

    Args:
        old_pixels (:obj:2D array): Input image data to be centred.
        min_pixel (int): Value to use for pixels outside the bounds of
            `old_pixels`.
        dim ((int, int)): Height and width of `old_pixels`.
        max_dim ((int, int), optional): Height and width of resized image.
            If set, this should be larger than `dim` in both dimensions.
            Defaults to maximum dimensions calculated in image dataset.
        dtype: (str, optional): Numpy-style data-type identifier.
            Defaults to 'float32'.

    Returns:
        numpy 2D array: New image, with the input image centred within it.

    """
    if max_dim is not None:
        max_dim = max_dim[:2]
    else:
        max_dim = get_hwd()[:2]
    new_pixels = np.empty(shape=max_dim, dtype=dtype)
    new_pixels[:,:] = min_pixel
    row_start = (max_dim[0] - dim[0]) // 2
    row_end = row_start + dim[0]
    col_start = (max_dim[1] - dim[1]) // 2
    col_end = col_start + dim[1]
    new_pixels[row_start:row_end, col_start:col_end] = old_pixels
    return new_pixels


def resize_pixels_to_max_dim(pixels, max_dim=None):
    """Return array of pixels containing the input array centred.

    The ouput 2D array has dimensions determined by either `max_dim` or
    the maximum dimensions calculated from the image dataset.

    Args:
        pixels (:obj:2D array): Input image data to be centred.
        max_dim ((int, int), optional): Height and width of resized image.
            If set, this should be larger than `dim` in both dimensions.
            Defaults to maximum dimensions calculated in image datset.

    Returns:
        numpy 2D array: New image, with the input image centred within it.

    """
    dim = (pixels.shape[0], pixels.shape[1])
    dtype = pixels.dtype
    min_pixel = np.min(pixels)
    new_pixels = resize_helper(pixels, min_pixel, dim, max_dim, dtype)
    return new_pixels


def resize_img_to_max_dim(img_name, max_dim=None):
    """Return array of pixels with the pixels representing `img_name` centred.

    Args:
        img_name (str): Fully-qualified path to a DICOM image.
        max_dim ((int, int), optional): Height and width of resized image.
            If set, this should be larger than the image's dimensions.
            Defaults to maximum dimensions calculated in the image dataset.

    Returns:
        numpy 2D array: Image, with the specified image centred within it.

    Raises:
        ValueError: If `img_name` does not point to a DICOM image.

    """
    header = pydicom.read_file(img_name)
    dim = (header.Rows, header.Columns)
    try:
        dtype = str(header.pixel_array.dtype)
    except ValueError:
        print("\nCannot read datatype of pixel array for DICOM image:")
        print(img_name)
        sys.exit(0)
    min_pixel = np.min(header.pixel_array)
    pixels = resize_helper(header.pixel_array, min_pixel, dim, max_dim, dtype)
    return pixels


def save_resized_img(img_name, pixels, max_dim):
    """Save pixels representing image to specified path.

    Args:
        img_name (str): Fully-qualified path to save the image at.
        pixels (:obj:2D array): Pixels representing an image.

    Returns:
        None

    """
    header = pydicom.read_file(img_name)
    os.chmod(img_name, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP |
                       stat.S_IROTH | stat.S_IXOTH
             )
    header.PixelData = pixels.tostring()
    header.Rows = max_dim[0]
    header.Columns = max_dim[1]
    header.save_as(img_name)


def resize_to_max_dim(max_dim=None):
    """Resize and save to disk all images in dataset.

    Args:
        max_dim ((int, int, int)): Height and width of resized image.
            If set, this should be larger than the image's dimensions.
            Defaults to maximum dimensions calculated in the image dataset.

    Returns:
        None

    """
    mouse_types = ["Anabolic", "Ovarectomy", "WildType"]
    if max_dim is not None:
        max_dim = max_dim[:2]
    else:
        max_dim = get_hwd()[:2]
    for t in mouse_types:
        for m in mouse_ids:
            for w in weeks:
                dirname = get_dir_name(t, m, w)
                if not os.path.exists(dirname):
                    continue
                img_names = os.listdir(dirname)
                os.chdir(dirname)
                for i in img_names:
                    pixels = resize_img_to_max_dim(i, max_dim=max_dim)
                    save_resized_img(i, pixels, max_dim)
    os.chdir(os.path.join(os.pardir, os.pardir))


def img_shift(img, row_shift, col_shift):
    """Translate an image by the specified number of rows and columns.

    Pixels which would be empty after the translation are set to the minimum
    pixel value observed in `img` prior to translation.
    This is based on the below, but with zero-value handling added:
    http://stackoverflow.com/questions/30399534/shift-elements-in-a-numpy-array#30534478

    Args:
        img (:obj:2D array): Pixel array to be translated.
        row_shift (`int`): y-axis translation, in pixels.
        col_shitf (`int`): x-axis translation, in pixels.

    Returns:
        numpy 2D array of same dimensions as `img`.

    """
    row_shift = int(row_shift)
    col_shift = int(col_shift)
    e = np.empty_like(img)
    img_min = np.min(img)
    e[:,:] = img_min
    if row_shift >= 0 and col_shift >= 0:
        e[row_shift:, col_shift:] = \
                img[:-row_shift if row_shift > 0 else img.shape[0],
                    :-col_shift if col_shift > 0 else img.shape[1]
                    ]
    elif row_shift >= 0 and col_shift < 0:
        e[row_shift:, :col_shift] = \
                img[:-row_shift if row_shift > 0 else img.shape[0],
                    -col_shift:
                    ]
    elif row_shift < 0 and col_shift >= 0:
        e[:row_shift, col_shift:] = \
                img[-row_shift:,
                    :-col_shift if col_shift > 0 else img.shape[1]
                    ]
    else:
        # As shifts are negative, swap negatives where we'd expect in here.
        e[:row_shift, :col_shift] = img[-row_shift:, -col_shift:]
    return e


def load_samples_by_week():
    """Return metadata about each image in the dataset.

    The metadata includes
    - the mouse type,
    - the week of the experiment in which the image data was scanned,
    - the directory on disk in which to find the images.

    Returns:
        `list` of tuples of (`int`, `int`, `str`)
            representing (treatment class, week, directory containing images).

    """
    train_weeks = []
    test_weeks = []

    for t in mouse_types:
        lbl = mouse_type_to_label(t)
        for m in mouse_ids:
            for w in weeks:
                dirname = get_dir_name(t, m, w)
                if not os.path.exists(dirname):
                    continue

                which = train_weeks if choice([1]*4 + [0]*1) else test_weeks
                
                print('Training:', t, m, w) \
                        if which == train_weeks \
                        else print('Testing:', t, m, w)

                for i in os.listdir(dirname):
                    d = os.path.join(dirname, i)
                    which.append( (lbl, w, d) )
    return (train_weeks, test_weeks)


def partition_samples(samples, all_mice=False):
    """Partition the provided list of `samples` into distinct groups.

    Samples are expected to correspond to a mouse-week of data, not an
    individual image data point.

    The samples will be divided into training, validation, and optionally also
    test partitions, which are all distinct from one another.
    The samples are randomly permuted before partitioning.

    If considering all mice, the training set will include 81% of samples,
    the validation set 9%, and the test set 10%.
    Otherwise, the training set will include 90% of samples and the validation
    set 10%.

    Args:
        samples (`list`): Tuples of (treatment class, week,
            image data directory) to be partitioned.
        all_mice (`bool`): Whether the samples include all mice or only ones
            used for training, hence whether or not to produce a test set.
            Defaults to True.

    Returns:
        `tuple` of (training, validation) if `all_mice` is `False` else
            (training, validation, testing).

    """
    samples = np.random.permutation(samples)
    val_split = 0.9
    l = samples.shape[0]
    if all_mice:
        train = samples[:int(val_split * val_split * l)]
        val = samples[int(val_split * val_split * l):int(val_split * l)]
        test = samples[int(val_split * l):]
        return (train, val, test)
    else:
        train = samples[:int(val_split * l)]
        val   = samples[int(val_split * l):]
        return (train, val)


def get_storage_files():
    """Return training, validation, and test dataset cached file locations.

    The training, validation, and test dataset samples are cached in numpy files
    on disk for speed. This returns the locations of those files.

    Args:
        None

    Returns:
        `tuple` of (training, validation, testing) dataset locations on disk.

    """
    train_file  = "train_labelled.npy"
    val_file    = "val_labelled.npy"
    test_file   = "test_labelled.npy"
    basedir = os.getcwd()
    while os.path.split(basedir)[1] != "Images":
        if "Images" in os.listdir(basedir):
            basedir = os.path.join(basedir, "Images")
            break
        basedir = os.path.split(basedir)[0]
    if not os.path.exists(basedir):
        os.mkdir(basedir)
    train_file  = os.path.join(basedir, train_file)
    val_file    = os.path.join(basedir, val_file)
    test_file   = os.path.join(basedir, test_file)
    return (train_file, val_file, test_file)


def load_partitioned_samples():
    """Return training, validation, and test dataset samples from cached files.

    Args:
        None

    Returns:
        `tuple` of (training, validation, testing) sample `list`s.

    """
    train_file, val_file, test_file = get_storage_files()
    train   = np.load(train_file)
    val     = np.load(val_file)
    test    = np.load(test_file)
    return (train, val, test)


def load_single(sample):
    """Return metadata and image data for a single sample.

    Args:
        sample (`tuple`): A single sample, containing a treatment class,
            week, and image data directory.

    Returns:
        `tuple` of (treatment class, week, pixel data)

    """
    header = pydicom.read_file(sample[2])
    rows = header.Rows
    cols = header.Columns
    pixels = np.reshape(header.pixel_array.astype("float32"), (rows, cols))
    return (int(sample[0]), float(sample[1]), pixels)


def load_to_global_by_week(use_diff=False, diff=''):
    global train_global
    global val_global
    global test_global

    if use_diff:
        train, test = load_diff_samples_by_week(diff)
    else:
        train, test = load_samples_by_week()

    train_global, val_global = partition_samples(train, all_mice=False)
    test_global = np.random.permutation(test)


def generate_helper(which, resize, all_mice=False):
    """Generate samples from the specified partition of samples.

    Args:
        which (`str`): Which set of samples to generate from.
            One of ["train", "val", "test"].
        resize (`bool`): Whether or not to perform resizing.
        all_mice (`bool`): Which sample partitions to load to global variables,
            if this has not already been done.
            If False, load only training and validation partitions,
            otherwise load test samples too.
            Defaults to True.

    Yields:
        `tuple` of (treatment class, week, image data).

    """
    data = None
    if train_global is None:
        raise ValueError('missing global data')
    if "train" == which.lower():
        data = train_global
    elif "val" == which.lower():
        data = val_global
    elif "test" == which.lower():
        data = test_global
    else:
        yield ValueError("`which` was not valid")
    while True:
        for sample in data:
            s = load_single(sample)
            if resize:
                s = (s[0], s[1], resize_pixels_to_max_dim(s[2]))
            yield (s)


def generate_mouse_data(which, resize, t=True, all_mice=False):
    """Generate a batch of data from the specified partition.

    Args:
        which (`str`): Which set of samples to generate from.
            One of ["train", "val", "test"].
        resize (`bool`): Whether or not to perform resizing.
            Defaults to True.
        t (`bool`): Whether or not to provide timestamps.
        all_mice (`bool`): Which sample partitions to consider.
            If False, consider only training and validation partitions.
            If True, additionally consider the test partition.
            Defaults to True.

    Yields:
        If `t` is True:
            {
                'input_1': `list` of timestamps,
                'input_2': `list` of pixel arrays,
                'dense_3': `list` of treatment classifications,
            }
        Else:
            {
                'input_1': `list` of pixel arrays,
                'dense_3': `list` of treatment classifications,
            }

    """
    global batch_size
    labels = np.empty((batch_size, len(mouse_types)), dtype="float32")
    times  = np.empty((batch_size,), dtype="float32")
    imgs   = np.empty((batch_size, channels, height, width), dtype=object)
    gen = generate_helper(which, resize, all_mice=all_mice)

    while True:
        for i in range(batch_size):
            sample = next(gen)
            labels[i] = np_utils.to_categorical(sample[0], len(mouse_types))
            times[i]  = sample[1]
            imgs[i]   = sample[2]
        if t:
            yield ({"input_1": times, "input_2": imgs}, {"dense_3": labels})
        else:
            yield ({"input_1": imgs}, {"dense_3": labels})


def load_diff_samples_by_week(diff):
    """Return samples containing metadata for use in differencing models.

    Args:
        diff (`str`): Which method of differencing to use.
            One of ["abs", "rel"].

    Returns:
        `list` of `tuple`s of (treatment class, timestamp,
            comparison image file location, reference image file location).

    """
    train_weeks = []
    test_weeks = []
    base_imgs = []

    for t in mouse_types:
        for m in mouse_ids:
            for w in weeks:
                dirname = get_dir_name(t, m, w)
                if not os.path.exists(dirname):
                    continue

                lbl = mouse_type_to_label(t)
                which = train_weeks if choice([1]*4 + [0]*1) else test_weeks
                img_names = os.listdir(dirname)

                print('Training:', t, m, w) \
                        if which == train_weeks \
                        else print('Testing:', t, m, w)

                if 0 == w:
                    base_imgs = [] # Reset - new mouse
                    for i in img_names:
                        f = os.path.join(dirname, i)
                        which.append( (lbl, w, f, f) )
                        base_imgs.append(f)
                else:
                    l = min(len(base_imgs), len(img_names))
                    for i in range(l):
                        base = base_imgs[i]
                        f = os.path.join(dirname, img_names[i])
                        which.append( (lbl, w, f, base) )
                    if 'rel' == diff.lower():
                        base_imgs = [
                                os.path.join(dirname, i)
                                for i in img_names
                                ]
    return (train_weeks, test_weeks)


def pixel_difference(img, diff_img):
    """Return a pointwise difference between two 2D arrays representing images.

    The input pixel arrays are first expanded to consistent dimensions
    and aligned.

    Args:
        img (:obj:2D array): Comparison image pixel array.
        diff_img (:obj:2D array): Reference image pixel array.

    Returns:
        numpy 2D array: Pointwise difference between expanded, aligned arrays.

    """
    shifts, _, _ = register_translation(diff_img,
                                        img,
                                        space="real",
                                        )
    diff = img_shift(img, shifts[0], shifts[1])
    diff -= diff_img
    return diff


def generate_diff_helper(which, resize, diff, all_mice=False):
    """Generate differencing samples from the specified partition of samples.

    Args:
        which (`str`): Which set of samples to generate from.
            One of ["train", "val", "test"].
        resize (`bool`): Whether or not to perform resizing.
        diff (`str`): Which method of differencing to use.
            One of ["abs", "rel"].
        all_mice (`bool`): Which sample partitions to load to global variables,
            if this has not already been done.
            If False, load only training and validation partitions,
            otherwise load test samples too.
            Defaults to True.

    Yields:
        `tuple` of (treatment class, week, comparison image data,
            reference image data).

    """
    data = None
    if train_global is None:
        raise ValueError('missing global data')
    if "train" == which.lower():
        data = train_global
    elif "val" == which.lower():
        data = val_global
    elif "test" == which.lower():
        data = test_global
    else:
        yield ValueError("`which` was not valid")
    while True:
        for sample in data:
            s = load_single(sample[0:3])
            sd = load_single([sample[0], sample[1], sample[3]])
            if resize:
                s = (s[0], s[1], resize_pixels_to_max_dim(s[2]))
                sd = (sd[0], sd[1], resize_pixels_to_max_dim(sd[2]))
            s = (s[0],
                 s[1],
                 s[2].astype("float32"),
                 pixel_difference(s[2], sd[2]).astype("float32")
                 )
            yield (s)


def generate_diff_mouse_data(which, resize, t, diff, all_mice=False):
    """Generate batches of image data and metadata for differencing models.

    Args:
        which (str): Which dataset to use - training, validation, or testing.
        resize (bool): Whether or not to perform resizing on loaded images.
        t (bool): Whether or not to include timestamps.
        diff (str): Type of differencing to use - 'abs' or 'rel'.

    Yields:
        If `t` is `True`:
            {
                input_1 (list): Timestamps.
                input_2 (list): Comparison images.
                input_3 (list): Reference images.
                dense_3 (list): Treatment classes.
            }
        Otherwise:
            {
                input_1 (list): Comparison images.
                input_2 (list): References images.
                dense_3 (list): Treatment classes.
            }

    """
    global batch_size
    global mouse_types
    height, width, channels = get_hwd()
    labels = np.empty((batch_size, len(mouse_types)), dtype="float32")
    times  = np.empty((batch_size,), dtype="float32")
    imgs   = np.empty((batch_size, channels, height, width), dtype=object)
    diffs  = np.empty((batch_size, channels, height, width), dtype=object)
    gen    = generate_diff_helper(which, resize, diff, all_mice=all_mice)

    while True:
        for i in range(batch_size):
            sample = next(gen)
            labels[i] = np_utils.to_categorical(sample[0], len(mouse_types))
            times[i]  = sample[1]
            imgs[i]   = sample[2]
            diffs[i]  = sample[3]
        if t:
            yield ({"input_1": times, "input_2": imgs, "input_3": diffs},
                   {"dense_3": labels}
                   )
        else:
            yield ({"input_1": imgs, "input_2": diffs}, {"dense_3": labels})

################################################################################

def test():
    # Test mouse_type_to_label.
    print(mouse_type_to_label("Anabolic"))
    print(mouse_type_to_label("Ovarectomy"))
    print(mouse_type_to_label("WildType"))
    try:
        print(mouse_type_to_label(""))
    except KeyError:
        pass
    # Test get_dir_name.
    get_dir_name("Anabolic", 4, 8)
    # Test get_max_dim.
    print(get_max_dim())
    # Test get_hwd.
    print(get_hwd())
    # Test resize_to_max_dim.
    resize_to_max_dim()
    print("Resize complete")
    # Test img_shift.
    img = pydicom.read_file("Anabolic/M1W0/WildPTHM1W0N1164Transformed0001.dcm")
    img = img.pixel_array
    shifted = img_shift(img, 10, 10)
    print("Image shift complete")

################################################################################

def main():
    get_max_dim()
    resize_to_max_dim()

################################################################################

if __name__ == "__main__":
    main()
