from datetime import datetime
import os
import sys

from tensorflow.python.lib.io import file_io

from . import parser
from .downloader import Downloader


def download_images(is_test=True):
    dl = Downloader(is_test=is_test)
    dl.download_files()


def get_log_dir(args):
    now = datetime.now().strftime('%Y%m%d_%H%M%S')

    cnn_type = 'xcnn' if args.is_xcnn else 'cnn'
    timestamp = '_t' if args.use_timestamps else ''
    difference = ''
    if args.is_xcnn:
        if args.use_abs_diff:
            difference = '_abs'
        else:
            difference = '_rel'

    dir_name = ''.join([cnn_type, timestamp, difference, '_', str(args.num_epochs), '_', now])
    return os.path.join(args.job_dir, dir_name)


def redirect_stdout(log_dir, args):
    log_file = os.path.join(log_dir, 'model.log')
    log_out = file_io.FileIO(log_file, 'w')
    sys.stdout = log_out


def run_model(log_dir, args, is_test=True):
    print('Dir: ', args.job_dir)
    print('Epochs: ', args.num_epochs)
    print('Batch size: ', args.batch_size)
    print('Is X-CNN: ', args.is_xcnn)
    print('Use timestamps: ', args.use_timestamps)
    print('Use absolute differencing: ', args.use_abs_diff)
    print('Use relative differencing: ', args.use_rel_diff)
    sys.stdout.flush()

    if args.is_xcnn and not args.use_timestamps \
            and not (args.use_abs_diff or args.use_rel_diff):
        raise ValueError('Must specify -t or -d for models with -x set')

    # Select model.
    if not args.is_xcnn:
        if args.use_timestamps:
            from .models.cnn_timestamps import CNNTimestamps as Model
        else:
            from .models.cnn_baseline import CNN as Model
    else:
        if args.use_timestamps:
            if args.use_abs_diff:
                from .models.xcnn_timestamps_and_absolute_differencing \
                        import CNNAbsoluteDifferencingTimestamps as Model
            else:
                from .models.xcnn_timestamps_and_relative_differencing \
                        import CNNRelativeDifferencingTimestamps as Model
        else:
            if args.use_abs_diff:
                from .models.xcnn_absolute_differencing \
                        import CNNAbsoluteDifferencing as Model
            else:
                from .models.xcnn_relative_differencing \
                        import CNNRelativeDifferencing as Model

    model = Model(
            log_dir=log_dir,
            batch_size=args.batch_size,
            num_epochs=args.num_epochs,
            is_test=is_test,
            )
    model.run()


if __name__ == '__main__':
    is_test = False
    args = parser.get_args()

    log_dir = get_log_dir(args)
    file_io.recursive_create_dir(log_dir)

    redirect_stdout(log_dir, args)

    download_images(is_test=is_test)

    run_model(log_dir, args, is_test)

    print('Completed; exiting')
    sys.stdout.flush()
